# -*- coding: utf-8 -*-
"""
Created on Wed Sep  8 09:55:41 2021
    Generates SMU offset and gain calibrations for 
    current Force (If) and Measure (Im) 
    Cal values are in output C format, to be copied to DIY-SMU cal.cpp file
    Each pass generates one range cal values. Should auto add them to a file
    Menu added 9/11/21. Allows one range to be selected. 
    Range setting uses dictionary rangelist for parameters
    Most ranges use 10V DMM range: 1mA 10K ohms, etc. 
    1uA uses DMM 1V range to allow 1Meg resstor. 10Meg precision R are tough
    100mA range uses 1V, 10 ohms. 10V would dissipate 1W, too much for precision R
    Cal algorithm:
        Prompt operator for range, and to install precision load resistor Rl
        Set range of force values, number of steps, - to + full scale
        Step thru force range, measure DMM V actoss Rl to get actual If
        Get Im from SMU
        Subtract measured from desired voltages to get errors.  
        Do linear curve fit to get Gain and Offset errors for I Force and I Meas
        Generate C format to be copied to DIY-SMU code
        
    Ultimately, cal values should be sent to SMU and loaded into EEPROM
    
@author: dave
"""

import time
import matplotlib.pyplot as plt
import numpy as np
import pyvisa
import datetime

# Globals and defines
#smuResource = 'ASRL4::INSTR'  # Unit 1
#smuResource = 'ASRL10::INSTR'  # Unit 2
smuResource = 'ASRL7::INSTR'  # Unit 4
# Test setup
measCount = 25
delayTime = 0.1
temp = 23.0

# Plot data 
dataDmm = []
dataForcErr = []
dataMeasErr = []
# Place to put the cal data :FG, FO, MG, MO  
calData = np.array([[1.0,0.0,1.0,0.0],[1.0,0.0,1.0,0.0],[1.0,0.0,1.0,0.0]])

rangeList = []              # Range data
rangeList.append(
        {'value':0.000001,'label':'1uA','res':1e6,
         'calLab':'IRange_1uA','dmmRange':'1, .000001'})
rangeList.append(
        {'value':0.00001,'label':'10uA','res':1e6,
         'calLab':'IRange_10uA','dmmRange':'10, .00001'})
rangeList.append(
        {'value':0.0001,'label':'100uA','res':100000.0,
         'calLab':'IRange_100uA','dmmRange':'10, .00001'})
rangeList.append(
        {'value':0.001,'label':'1mA','res':10000.0,
         'calLab':'IRange_1mA','dmmRange':'10, .00001'})
rangeList.append(
        {'value':0.01,'label':'10mA','res':1000.0,
         'calLab':'IRange_10mA','dmmRange':'10, .00001'})
rangeList.append(
        {'value':0.1,'label':'100mA','res':10.0,
         'calLab':'IRange_100mA','dmmRange':'1, .000001'})


def menu():
    print("Enter range: \n0: Quit, 1: 1uA, 2: 10uA, 3: 100uA, 4: 1mA, 5: 10mA 6: 100mA ")

def plot(x, datalist, plotLabel):
    # plot datalist
    plt.figure(figsize=(8,4))
#    plt.plot(x, datalist, label='DMM vs SMU error, VDC')
    plt.plot(x, datalist, label=plotLabel)
    plt.xlabel('Force I')
    plt.ylabel('Error: Current')
    plt.title("SMU Current Calibration")
    plt.legend()
    plt.grid(True)
    plt.show()

def initVISA():
    print("Setting up GPIB")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    return rm

def initSMU():    # DIY-SMU setup SMU
    #1 is ASLR9, SMU2 is ASLR10
    print("Setting up DIY-SMU")
    #smu = rm.open_resource('ASRL10::INSTR')
    smu = rm.open_resource(smuResource)
    smu.read_termination  = '\n'            #needed for SMU
    smu.write_termination = '\n'
    print(smu.query("*IDN?"))
    smu.write("CAL:STAT ON")
    smu.write("FUNC:CURR")
    smu.write("SOUR:VOLT:RANG 15") 
    return smu

def initDMM():
    # Get Data from DMM HP34401A GPIB addr 23
    dmm = rm.open_resource('GPIB0::23::INSTR')
    print(dmm.query("*IDN?"))
    #dmm.write("*RST")
    #time.sleep(3)
    return dmm
    
def calCurrent(rangeNum):
    index = 0      
    instrRange = rangeList[rangeNum]["value"]
    calR = rangeList[rangeNum]["res"]
    rangeName = str(instrRange)
    print("rangeName:", rangeName)
    print ("Cal Resistor = ", calR)
    smu.write("SOUR:CURR:RANG " + rangeName)  
    smu.write("OUTP:STAT ON")
    startI = -1.0 * instrRange  # Set the force current values
    endI =    1.0 * instrRange 
    print("Calibrating range:", instrRange)
    print("From", startI, "to", endI)
    force_i = np.linspace(startI, endI, measCount + 1)
    dataDmm = []
    dataForcErr = []
    dataMeasErr = []
    time.sleep(5*delayTime)    #Delay to setle range change and output ON
   
    if instrRange == .000001 or instrRange == .1 : # use 1V on these ranges
        dmm.write("CONF:VOLT:DC 1, 0.000001") 
    else:
        dmm.write("CONF:VOLT:DC 10, 0.00001")
    
    dmm.write("INP:IMP:AUTO ON") # Put DMM in hi-z mode
#       dmm.write("TRIG:SOUR BUS")
    for count in range(0, np.size(force_i)):   
        forceCurr = force_i[count]
        smu.write("SOUR:CURR " + str(forceCurr))
        time.sleep(delayTime)
 #       dmm.write("MEAS:VOLT:DC?") #Enable for autoscale, disable others
 #       dmm.write("MEAS:VOLT:DC? 10, 0.00001")

        dmm.write("READ?")
        dmmMeasV = float(dmm.read()) #get dmm data first since it is synched 
        dmmMeasA = dmmMeasV / calR   #convert to current
        smu.write("MEAS:CURR?") 
        smuMeasA = float(smu.read()) 
        print ("Fcurr", forceCurr)
#        print ("DMM curr, V/R", dmmMeasA)
#        print ("SMU Meas I", smuMeasA)
        forceErr = forceCurr - dmmMeasA
        measErr  = smuMeasA  - dmmMeasA 
#        print ("IF Err",  forceErr)
#        print ("IM Err",  measErr)
        dataDmm = dataDmm + [dmmMeasA]
        dataForcErr = dataForcErr + [forceErr]
        dataMeasErr = dataMeasErr + [measErr] 
    
    linForce = np.polyfit(force_i, dataForcErr, 1)
    calData[index, 0] = 1 - linForce[0]
    calData[index, 1] = -linForce[1]
    
#    print ("Force Error Coeff:" , linForce)
    print ("Force GAIN Cal:", 1 - linForce[0])
    print ("Force Offset Cal:", -linForce[1])
    
    
    linMeas = np.polyfit(dataDmm, dataMeasErr, 1) # For VM, use DMM as ref
    calData[index, 2] = 1 + linMeas[0]
    calData[index, 3] = linMeas[1]
#    print ("Meas Error Coeff:", linMeas)
    print ("Meas GAIN Cal:", 1 + linMeas[0])
    print ("Meas Offset Cal:", linMeas[1])
    
    print("Min:", startI, "Max:", endI)
    #print("Force Error")
    plot(force_i,dataForcErr, "Force V Error")
    #print("Measure Error")  
    plot(force_i,dataMeasErr, "Measure V Error")
    
    index += 1
        
    # Print C format cal factors in the format SMU wants
    print("\n\n")
    index = 0
    print("// Calibrated " + str(datetime.datetime.now()) + " Temp.: " + str(temp))

    # Print out the cal data in C language format that DIY-SMU wants
    rangeName = str(instrRange)
    smuRange = rangeList[rangeNum]["calLab"]
    print("//", rangeName, "I Range Cal")
    print("Range[" + smuRange + "].FGainCal = %11.8f ;" % calData[index, 0])
    print("Range[" + smuRange + "].FOffsCal = %11.8f ;" % calData[index, 1])
    print("Range[" + smuRange + "].MGainCal = %11.8f ;" % calData[index, 2])
    print("Range[" + smuRange + "].MOffsCal = %11.8f ;" % calData[index, 3])

    index += 1     
    
rm = initVISA()
dmm = initDMM()
smu = initSMU()

print ("DIY-SMU Current range calibration")
    
menu()

option = int(input("Enter your Option:"))

while option != 0:
    if (option > -1 and option < 7 ):
        rangeNum = option-1
        print("Install cal resistor: " + str(rangeList[rangeNum]["res"]) + " Ohms and press Enter.")
        input("")
        print ("Doing cal of " + rangeList[rangeNum]["label"] + " range") # menu is 1 based, list is 0 based
        calCurrent(rangeNum)
        print ("/nCal Complete")
        
        
    else:
        print ("Incorrect range")
        
    print()
    menu()
    option = int(input("Enter your Option:"))

# Need to close instruments when done. If program hangs, do this manually
dmm.close()
smu.close()
print ("Bye!")
