"""
 Voltage Calibration code for SMU
 Measures SMU force using DMM and SMU Measure, then Generates cal factors
 GPIB uses  HP34401A GPIB Addr: 23
 
 Requires NI GPIB, NI Visa, pyvisa, numpy, matplotlib 

 Make sure all instruments are on, USB / GPIB connected, and cabled up.
 Make sure DMM GPIB address and SMU Serial port are correct

 To generate cal factors, turn cal OFF: smu.write("CAL:STAT OFF")
 To verify cal accuracy, turn cal ON:   smu.write("CAL:STAT ON")
 Uses +/- 11V (not +/- 15V) since 34401A DMM maximum is +/- 12V on 10V range. 
 This permits the maximum 6 digit) resolution. 
 To exceed +/- 11V, the DMM must be set to Auto range or one voltage range up. 
    Generates SMU offset and gain calibrations for 
    voltage Force (Vf) and Measure (Vm) 
    Cal values are in output C format, to be copied to DIY-SMU cal.cpp file
    Each pass generates three range cal values. Should auto add them to a file
    Range setting uses dictionary rangelist for parameters
    Most ranges use 10V DMM range 

 Cal algorithm:
    No prompt needed, calibrates all 3 voltage ranges
    Set range of V force values, number of steps, - to + full scale
    Step thru force range, measure DMM V to get actual Vf
    Get Vm from SMU
    Subtract measured from desired voltages to get errors.  
    Do linear curve fit to get Gain and Offset errors for V Force and Meas
    Generate C format to be copied to DIY-SMU code
    
Ultimately, cal values should be sent to SMU and loaded into EEPROM
"""

#import serial
import time
import matplotlib.pyplot as plt
import numpy as np
import pyvisa
import datetime

# Measure error
# Test setup
measCount = 20
delayTime = 0.25
temp = 22.0

#smuResource = 'ASRL4::INSTR'  # Unit 1
smuResource = 'ASRL10::INSTR' # Unit 3
#smuResource = 'ASRL7::INSTR'  # Unit 4


# Plot data 
dataDmm = []
dataForcErr = []
dataMeasErr = []
# Place to put the cal data for 3 ranges:FG, FO, MG, MO  
calData = np.array([[1.0,0.0,1.0,0.0],[1.0,0.0,1.0,0.0],[1.0,0.0,1.0,0.0]])

# To plot to a separate window with plot controls, type #
#       %matplotlib auto
# To plot inline type 
#       %matplotlib inline

def plot(x, datalist, plotLabel):
    # plot datalist
    plt.figure(figsize=(8,4))
#    plt.plot(x, datalist, label='DMM vs SMU error, VDC')
    plt.plot(x, datalist, label=plotLabel)
    plt.xlabel('Force V')
    plt.ylabel('Error (Volts)')
    plt.title("SMU Voltage Calibration")
    plt.legend()
    plt.grid(True)
    plt.show()


print("Setting up GPIB")
rm = pyvisa.ResourceManager()
print(rm.list_resources())

# DIY-SMU setup SMU#1 is ASLR9, SMU2 is ASLR10
print("Setting up DIY-SMU")
#smu = rm.open_resource('ASRL10::INSTR')
smu = rm.open_resource(smuResource)
smu.read_termination  = '\n'            #needed for SMU
smu.write_termination = '\n'

print(smu.query("*IDN?"))

# Get Data from DMM HP34401A GPIB addr 23
# 34401 10V range measures 12V MAX, so need 100V range or auto to measure +/- 15V
dmm = rm.open_resource('GPIB0::23::INSTR')
print(dmm.query("*IDN?"))

smu.write("CAL:STAT OFF")   #OFF  for calibration, ON for verification
smu.write("FUNC:VOLT")      # Voltage mode

   
#vRange = [15]              # cal one range
vRange = [1.5, 15, 150]     # cal all ranges
index = 0

for instrRange in vRange:
    rangeName = str(instrRange)
#    print("rangeName:", rangeName)
    smu.write("SOUR:CURR:RANG .001")            # 1mA current range
    smu.write("SOUR:VOLT:RANG " + rangeName)  
    smu.write("OUTP:STAT ON")
    startV = -15 * instrRange / 15 
    endV =    15 * instrRange / 15 
    print("Calibrating range:", instrRange)
    print("From", startV, "to", endV)
    
    force_v = np.linspace(startV, endV, measCount + 1)
    dataDmm = []
    dataForcErr = []
    dataMeasErr = []
    time.sleep(delayTime)    #Delay to setle range change and output ON
    
    for count in range(0, np.size(force_v)):   
        forceVolt = force_v[count]
        smu.write("SOUR:VOLT " + str(forceVolt))
        time.sleep(delayTime)
#        dmm.write("MEAS:VOLT:DC?") #Enable for autoscale, disable others
        if instrRange == 1.5:
            dmm.write("MEAS:VOLT:DC? 10, 0.00001")
        if instrRange == 15:
            dmm.write("MEAS:VOLT:DC? 100, 0.0001")
        if instrRange == 150:
            dmm.write("MEAS:VOLT:DC? 1000, 0.001")
        dmmMeas = float(dmm.read()) #get dmm data first since it is synched 
        smu.write("MEAS:VOLT?") 
        smuMeas = float(smu.read()) 
        print ("Fvolt %10.6f" % forceVolt)
    #    print ("dmmData ", dmmMeas)
    #    print ("smuMeas ", smuMeas)
        forceErr = forceVolt - dmmMeas
        measErr  = smuMeas - dmmMeas
    #    print ("VF Err: %11.8f " % forceErr)
    #    print ("VM Err: %11.8f " % measErr)
        dataDmm = dataDmm + [dmmMeas]
        dataForcErr = dataForcErr + [forceErr]
        dataMeasErr = dataMeasErr + [measErr] 
        smu.write("SOUR:VOLT 0.0")
    
    linForce = np.polyfit(force_v, dataForcErr, 1)
    calData[index, 0] = 1 - linForce[0]
    calData[index, 1] = -linForce[1]
    
    print ("Force Error Coeff: " , linForce)
    print ("Force GAIN Cal: ", 1 - linForce[0])
    print ("Force Offset Cal: ", -linForce[1])
    
    
    linMeas = np.polyfit(dataDmm, dataMeasErr, 1) # For VM, use DMM as ref
    calData[index, 2] = 1 + linMeas[0]
    calData[index, 3] = linMeas[1]
    print ("Meas Error Coeff:", linMeas)
    print ("Meas GAIN Cal: ", 1 + linMeas[0])
    print ("Meas Offset Cal: ", linMeas[1])
    
    print("Min:", startV, "Max:", endV)
    #print("Force Error")
    plot(force_v,dataForcErr, "Force V Error")
    #print("Measure Error")  
    plot(force_v,dataMeasErr, "Measure V Error")
    
    index += 1
    
# Print C format cal factors in the format SMU wants
print("\n\n")
index = 0
print("// Calibrated " + str(datetime.datetime.now()) + " Temp.: " + str(temp))

for instrRange in vRange:
    rangeName = str(instrRange)
    if instrRange == 1.5:
        smuRange= "VRange_1_5"
    if instrRange == 15:
        smuRange= "VRange_15 "
    if instrRange == 150:
        smuRange= "VRange_150"
    print("//", rangeName, "V Range Cal")
    print("Range[" + smuRange + "].FGainCal = %11.8f ;" % calData[index, 0])
    print("Range[" + smuRange + "].FOffsCal = %11.8f ;" % calData[index, 1])
    print("Range[" + smuRange + "].MGainCal = %11.8f ;" % calData[index, 2])
    print("Range[" + smuRange + "].MOffsCal = %11.8f ;" % calData[index, 3])

    index += 1


smu.write("CAL:STAT ON")       #Turn CAL ON    
# Need to close instruments when done. If program hangs, do this manually
dmm.close()
smu.close()




# useful junk and sample code
#print("Geeks : %2d, Portal : %5.6f" % (1, 05.1234567890))

#for index in range(len(values)):
#    value = values[index]
#    print(index, value)

























